package NCSU::RESTclient;

use warnings;
use strict;
use Carp;

use version; our $VERSION = qv('0.6.2');

# exports
use base qw( Exporter );
our @EXPORT      = qw( );
our @EXPORT_OK   = qw( );
our %EXPORT_TAGS = (
    ALL => [ @EXPORT, @EXPORT_OK ],    # if :ALL requested
);

# other dependent libraries
use Date::Format qw(time2str);
use Digest::SHA qw(hmac_sha1_base64 hmac_sha256_base64 hmac_sha384_base64);
use Digest::MD5;
use IO::Socket::SSL;                   # to check SSL handshake: qw(debug3);
use LWP;
use LWP::Protocol::https;
use MIME::Base64;
use Module::Load::Conditional qw(can_load);
use URI;

# Module implementation here

# create the module object
sub new {
    my ( $class, $paramref ) = @_;

    # start with defaults
    my $self = {
        apiver    => 'v1',
        authdata  => undef,
        authtype  => undef,
        baseurl   => undef,
        debug     => undef,
        encoder   => 'www',
        mediatype => 'json',
        sha       => 'sha256',
        timeout   => 30,
    };
    bless( $self, ref($class) || $class );

    # set defined params over defaults
    foreach my $attr ( keys %$self ) {
        $self->{$attr} = $paramref->{$attr} if ( defined $paramref->{$attr} );
    }

    return ($self);
}

# set authorization method and data
sub set_auth {
    my $self = shift;
    my $type = shift;
    my $data = shift;

    if ( $type eq 'statickey' ) {
        $self->{authtype} = $type;
        if ( defined $data->{keyid} ) {
            $self->{authdata}->{keyid} = $data->{keyid};
        }
        else {
            croak("missing keyid for static key");
        }
        if ( defined $data->{keydata} ) {
            $self->{authdata}->{keydata} = $data->{keydata};
        }
        else {
            croak("missing keydata for static key");
        }
    }
    elsif ( $type eq 'basic' ) {
        $self->{authtype} = $type;
        if ( defined $data->{username} ) {
            $self->{authdata}->{username} = $data->{username};
        }
        else {
            croak("missing username for basic auth");
        }
        if ( defined $data->{password} ) {
            $self->{authdata}->{password} = $data->{password};
        }
        else {
            croak("missing password for basic auth");
        }
    }
    else {
        croak("unknown authorization type '$type'");
    }

    return $self->{authtype};
}

# format and make an LWP request to a REST service
sub request {
    my $self    = shift;
    my $method  = shift;
    my $path    = shift;
    my $content = shift || undef;
    my $headers = shift || undef;

    # check request params
    croak("baseurl not set in object") if ( !$self->{baseurl} );
    my %allowed = map { $_ => 1 } qw(GET PUT POST DELETE PATCH);
    croak("Unknown method '$method'") if ( !$allowed{$method} );
    croak("URL path not passed") if ( !$path );

    # start building the request
    my $ua  = $self->_getuseragent();
    my $URL = URI->new( $self->{baseurl} . $path );
    my $req = HTTP::Request->new( $method, $URL );

    # pass any given headers along
    if ($headers) {
        if ( ref($headers) eq 'HASH' ) {
            foreach my $k ( keys %$headers ) {
                $req->header( $k => $headers->{$k} );
            }
        }
        else {
            croak("headers not passed as a hashref");
        }
    }

    # store the request in the object for debugging
    $self->{_request} = $req;

    # generate Date header
    my $date = time2str( '%a, %d %b %Y %T %Z', time, 'GMT' );
    $req->header( 'Date' => $date );

    # generate content headers, if we have content
    my $contentMD5 = '';
    if ( defined $content ) {

        if ( ref $content ) {
            if ( $self->{encoder} eq 'json' ) {
                $content = _encode_json($content);
            }
            else {
                # use this trick from HTTP/Request/Common to encode content
                my $url = URI->new('http:');
                $url->query_form($content);
                $content = $url->query;
                $content =~ s/(?<!%0D)%0A/%0D%0A/g;
            }
        }
        my $md5 = Digest::MD5->new;
        $md5->add($content);
        $contentMD5 = $md5->b64digest;
        $req->content($content);
        if ( $self->{encoder} eq 'json' ) {
            $req->content_type('application/json');
        }
        else {
            $req->content_type('application/x-www-form-urlencoded');
        }
        $req->header( 'Content-Length' => length($content) );
        $req->header( 'Content-MD5'    => $contentMD5 );
    }

    # generate the accept header for the requested version/format
    my $accept =
      'application/vendor.api-' . $self->{apiver} . '+' . $self->{mediatype};
    $req->header( 'Accept' => $accept );

    # try to generate the NCSU-MAC signature, if requested
    if ( $self->{authtype} eq 'statickey' ) {
        my $adata = $self->{authdata};
        if (   defined $adata
            && defined $adata->{keyid}
            && defined $adata->{keydata} )
        {
            my $tosign = join( "\n", $method, $path, $date, $contentMD5 );
            my $digest =
              $self->{sha} eq 'sha1'
              ? hmac_sha1_base64( $tosign, $adata->{keydata} )
              : $self->{sha} eq 'sha384'
              ? hmac_sha384_base64( $tosign, $adata->{keydata} )
              : hmac_sha256_base64( $tosign, $adata->{keydata} );
            my $sign = $adata->{keyid} . ':' . $digest;
            $req->header( 'NCSU-MAC' => $sign );
        }
        elsif ( $self->{debug} ) {
            carp("missing keyid and/or keydata, not sending NCSU-MAC");
        }
    }
    elsif ( $self->{authtype} eq 'basic' ) {
        my $adata = $self->{authdata};
        if (   defined $adata
            && defined $adata->{username}
            && defined $adata->{password} )
        {
            my $userpass =
              encode_base64( $adata->{username} . ':' . $adata->{password} );
            $req->header( 'Authorization' => "Basic $userpass" );
        }
        elsif ( $self->{debug} ) {
            carp("missing username and/or password, not sending Authorization");
        }
    }

    # execute the LWP request
    my $res = $ua->request($req);
    $self->{_response} = $res;
    if ( $res->is_success ) {
        return $res->content;
    }
    elsif ( $self->{debug} ) {
        carp( $res->status_line . "\n" . $res->content . "\n" );
    }
    return;
}

# decode controller
sub decode {
    my $self    = shift;
    my $content = shift || $self->content;
    return if ( !$content );

    return (
          $self->{mediatype} eq 'json' ? _decode_json($content)
        : $self->{mediatype} eq 'xml'  ? _decode_xml($content)
        : $self->{mediatype} eq 'yaml' ? _decode_yaml($content)
        :   croak("cannot decode unknown mediatype '$self->{mediatype}'")
    );
}

# passthru convenient functions for response object
sub code {
    my $self = shift;
    return if ( !defined $self->{_response} );
    return $self->{_response}->code;
}

sub content {
    my $self = shift;
    return if ( !defined $self->{_response} );
    return $self->{_response}->content;
}

sub is_success {
    my $self = shift;
    return if ( !defined $self->{_response} );
    return $self->{_response}->is_success;
}

sub status_line {
    my $self = shift;
    return if ( !defined $self->{_response} );
    return $self->{_response}->status_line;
}

# internal routine to initialize the LWP User Agent
sub _getuseragent {
    my $self = shift;

    return $self->{_ua} if ( defined $self->{_ua} );
    my $ua = LWP::UserAgent->new( agent => 'NCSU-RESTclient/' . $VERSION );
    $ua->timeout( $self->{timeout} ) if ( $self->{timeout} > 0 );
    $ua->cookie_jar( {} );    # give agent a temp cookie jar
    if ( $self->{debug} ) {
        $ua->add_handler( "request_send",
            sub { shift->dump; print "\n\n"; return } );
    }
    $self->{_ua} = $ua;
    return $ua;
}

# internal encoding routine for json
sub _encode_json {
    my $content = shift;
    if ( !can_load( modules => { 'JSON' => 2.0, } ) ) {
        croak("Failed to load JSON encoder library");
    }
    my $json = JSON->new->utf8(1);
    eval { return $json->encode($content) } or return;
}

# internal decoding routine for json
sub _decode_json {
    my $content = shift;
    if ( !can_load( modules => { 'JSON' => 2.0, } ) ) {
        croak("Failed to load JSON decoder library");
    }
    my $json = JSON->new->utf8(1);
    eval { return $json->decode($content) } or return;
}

# internal decoding routine for xml
sub _decode_xml {
    my $content = shift;
    if ( !can_load( modules => { 'XML::Simple' => 2.0, } ) ) {
        croak("Failed to load XML::Simple decoder library");
    }
    my $xs = XML::Simple->new();
    eval { return $xs->XMLin($content) } or return;
}

# internal decoding routine for yaml
sub _decode_yaml {
    my $content = shift;
    if ( !can_load( modules => { 'YAML::Loader' => 0, } ) ) {
        croak("Failed to load YAML::Loader decoder library");
    }
    my $yaml = YAML::Loader->new();
    eval { return $yaml->load($content) } or return;
}

1;    # Magic true value required at end of module
__END__

=head1 NAME

NCSU::RESTclient - extensible library for NCSU REST services


=head1 VERSION

This document describes NCSU::RESTclient version 0.6.1


=head1 SYNOPSIS

    use NCSU::RESTclient;

    my $client = NCSU::RESTclient->new( {
       baseurl => 'https://api.oit.ncsu.edu',
    } );
    $client->set_auth( 'statickey', {
       keyid   => 'idstr',
       keydata => 'longkeydatastring',
    } );

    $client->request('GET', '/users/jqpublic');
    die $client->status_line if (! $client->is_success);

    my $userdata = $client->decode();

  
=head1 DESCRIPTION

NCSU::RESTclient provides a simple, standard way to use perl to interact 
with standardized RESTful services on NCSU campus. 

The current version supports authorized requests using:
- NCSU-HMAC = a static key HMAC signature algorithm,
- BasicAuth 
It can also send unauthorized requests.

OAuth 1.0a (two-legged) and OAuth 2.0 authorization may be added in 
the future.


=head1 INTERFACE 

=head2 new

 Usage     : my $client = NCSU::RESTclient->new( { key => val, ... } );
 Purpose   : Initialize the object
 Returns   : A blessed object
 Arguments : hashref of key => value pairs including:
           : baseurl   => base URL of REST data application, required
           : apiver    => API version string to use, default 'v1'
           : debug     => set true for extra debugging messages
           : encoder   => encoder for sent content, default 'www'
           : mediatype => desired result format, default 'json'
           : sha       => 'sha1', 'sha256', or 'sha384', default 'sha256'
           : timeout   => timeout in seconds for the client, default 30

=head2 set_auth

 Usage:    : $client->set_auth($type, $hashref);
 Purpose   : Set the authorization type and key data for requests
 Returns   : The $type that was set. Will croak on failure.
 Arguments : $type    = string indicating the authorization type
           : $hashref = reference to hash of key data for the auth

 For static key HMAC authorization, use:
 $type = 'statickey'
 $hashref = { keyid => 'string', keydata => 'string' }

 For BasicAuth authorization, use:
 $type = 'basic'
 $hashref = { username => 'string', password => 'string' }

 For no authorization, omit this call.

=head2 request

 Usage     : my $result = $client->request( $method, $path, $content, $headers );
 Purpose   : Send a REST request with authorization
 Returns   : Encoded result content or undef on failure
 Arguments : $method  = one of qw( GET PUT POST DELETE PATCH )
           : $path    = URL path under baseurl, including any query data
           : $content = string of encoded content data for POST / PUT
           : $headers = optional hashref for additional headers

 $content can be a bare hashref or arrayref, and the content will be encoded 
 for you using the URI library.

=head2 decode

 Usage     : my $data = $client->decode($content);
 Purpose   : Decode content based on the expected mediatype
 Returns   : A perl object of the decoded data
 Arguments : $content = string to decode, or use the last response content
 Comments  : Assumes the object mediatype param is the correct format
           : json decode requires JSON module
           : xml  decode requires XML::Simple module
           : yaml decode requires YAML::Loader module
           : any other mediatype will fail

=head2 code

=head2 content

=head2 is_success

=head2 status_line

 Usage     : my $result = $client->content if ($client->is_success);
 Purpose   : Pass-thru HTTP::Response methods for convenience
 Returns   : HTTP::Response method results
 Comment   : These can only be used after a call to request() 


=head1 DIAGNOSTICS

=over

=item C<< missing keyid for static key >>

The set_auth call for type = 'statickey' did not include a keyid entry in 
the data.

=item C<< missing keydata for static key >>

The set_auth call for type = 'statickey' did not include a keydata entry
in the data.

=item C<< unknown authorization type '%s' >>

The set_auth call requested an auth type that is not recognized.
Known types are: 'statickey'.

=item C<< baseurl not set in object >>

The request call cannot be made because the baseurl attribute was not 
set in the object. You must set baseurl when new() is called.

=item C<< Unknown method '%s' >>

The request call must use one of these known methods: GET, PUT, POST,
DELETE, and PATCH. The method must be specified in UPPERCASE.

=item C<< URL path not passed >>

The request call must include the path of the request, relative to the
baseurl set in the object. 

=item C<< headers not passed as a hashref >>

If you pass additional headers to a request call, you must provide
those headers as a hashref of 'Header' => 'Value' pairs.

=item C<< missing keyid and/or keydata, not sending NCSU-MAC >>

If the object does not contain keyid and keydata attributes, then
the REST request will still be made but the NCSU-MAC auth header will
not be passed. The request will fail for REST services that require
authorization. This is a warning message that only appears when
debug is set. This error message should not appear if set_auth was 
called correctly.

=item C<< any HTTP error status line >>

If the REST request failed and debug is set, then resulting HTTP 
status message and any error message content will be displayed as 
a warning.

=item C<< cannot decode unknown mediatype '%s' >>

The decode call can only decode these mediatypes: json, xml, and yaml.

=item C<< Failed to load JSON encoder library >>

The encoder was set to 'json', but the JSON module was not found 
at runtime. The encode failed.

=item C<< Failed to load JSON decoder library >>

The JSON module was not found at runtime, so the decode failed.

=item C<< Failed to load XML::Simple decoder library >>

The XML::Simple module was not found at runtime, so the decode failed.

=item C<< Failed to load YAML::Loader decoder library >>

The YAML::Loader module was not found at runtime, so the decode failed.

=back


=head1 CONFIGURATION AND ENVIRONMENT

NCSU::RESTclient requires no configuration files or environment variables.


=head1 DEPENDENCIES

=over

=item Date::Format

=item Digest::SHA

=item Digest::MD5

=item IO::Socket::SSL - for proper hostname checking

=item LWP

=item Module::Load::Conditional

=item LWP::Protocol::https - for SSL-based REST URLs

=item URI

=item (optional) JSON - for json decoding

=item (optional) XML::Simple - for xml decoding

=item (optional) YAML::Loader - for yaml decoding

=back


=head1 INCOMPATIBILITIES

None reported.


=head1 BUGS AND LIMITATIONS

No bugs have been reported.

Please report any bugs or feature requests to the author.


=head1 AUTHOR

Charles J. Brabec  C<< <brabec@ncsu.edu> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2013, Charles J. Brabec C<< <brabec@ncsu.edu> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
